import os
import uuid
from typing import Dict, Any, List
import sys
import subprocess
import pandas as pd
import time
from datetime import datetime
from dateutil.parser import parse
import json
import pytz
import logging
from enum import Enum
from pyspark.sql import SparkSession
from pyspark.sql.functions import sum, collect_list, round, when, col, lit, size, udf, array_except, array
from pyspark.sql.types import ArrayType, IntegerType, LongType


class FixedFieldsEnum(Enum):
    INTER_PIVOT_ID = "INTER_PIVOT_ID"
    INTER_CTP_ID = "INTER_CTP_ID"
    LISTA_DIFF = "LISTA_DIFF"
    DIFF = "DIFF"
    MATCH_RECORDS = "match-records"


class StatusEnum(Enum):
    OK = 200
    ERROR = 609
    TIMEOUT = 610


logger = logging.getLogger(__name__)


# EXCLUDE VALIDATION FIELD
EXCLUDE_ROWS_FIELD = "EXCLUDE_VALID"

# REDONDEO DE DECIMALES
ROUND_DECIMAL = 2

# COLUMNAS TABLA RESULTADO
RESULT_TABLENAME = "CSS_RESULT_BY_ACTION"
RESULT_TABLE_FIELDS = ["ACTION_ID", "ID_PROCESS", "CREATE_DATE", "KEY", "RESULT_JSON"]


def process() -> Dict[str, Any]:
    response = {"status": StatusEnum.ERROR.name}
    start_time = time.time()
    params = sys.argv
    descriptor = params[1]
    jdbc_url = params[2]
    timezone = params[3]
    pattern = params[4]
    descriptor = json.loads(descriptor)

    session = createSession()
    configs = descriptor["config-params"]
    exclude_pivot = configs["exclude-entity-pivot"]
    max_combinations = configs["max-records-per-combinations"]

    params_input = descriptor["params-input"]
    pivot_params, ctp_params = params_input["pivot-config"], params_input["counterpart-config"]

    pivot_table, ctp_table = pivot_params["tablename"], ctp_params["tablename"]
    jdbc_properties = {"driver": "com.mysql.jdbc.Driver"}
    pivot_df = session.read.jdbc(url=jdbc_url, table=pivot_table, properties=jdbc_properties)
    ctp_df = session.read.jdbc(url=jdbc_url, table=ctp_table, properties=jdbc_properties)

    # Agregar un prefijo a cada columna, tanto del pivot como contraparte. Actualizar campos del input
    # pivot: 'PIVOT_', contraparte: 'COUNTERPART_'
    for column in pivot_df.columns:
        if column == EXCLUDE_ROWS_FIELD:
            continue
        pivot_df = pivot_df.withColumnRenamed(column, "PIVOT_" + column)

    for column in ctp_df.columns:
        if column == EXCLUDE_ROWS_FIELD:
            continue
        ctp_df = ctp_df.withColumnRenamed(column, "COUNTERPART_" + column)

    for key_p, key_c in zip(pivot_params.keys(), ctp_params.keys()):
        if isinstance(pivot_params[key_p], str):
            pivot_params[key_p] = "PIVOT_" + pivot_params[key_p]
            ctp_params[key_c] = "COUNTERPART_" + ctp_params[key_c]
        else:
            pivot_params[key_p] = ["PIVOT_" + column for column in pivot_params[key_p]]
            ctp_params[key_c] = ["COUNTERPART_" + column for column in ctp_params[key_c]]

    pivot_cols = pivot_params["columns-transaction"].copy()
    if pivot_params["amount-column"] in pivot_cols:
        pivot_cols.remove(pivot_params["amount-column"])

    ctp_cols = ctp_params["columns-transaction"].copy()
    if ctp_params["amount-column"] in ctp_cols:
        ctp_cols.remove(ctp_params["amount-column"])

    # Ejecutamos lógica de excluir registros
    if len(pivot_params["columns-group"]) == 0 and len(ctp_params["columns-group"]) == 0:
        raise RuntimeError(f"Debe haber al menos pivot o contraparte agrupado")

    # Caso: 1 - Muchos
    elif len(pivot_params["columns-group"]) == 0 and len(ctp_params["columns-group"]) > 0:
        ctp_df2 = ctp_df.groupby(ctp_params["columns-group"]). \
            agg(round(sum(ctp_params["amount-column"]), ROUND_DECIMAL).alias(ctp_params["amount-column"]),
                collect_list(ctp_params["id-column"]).alias(ctp_params["id-column"]))
        pivot_df2 = pivot_df
    # Caso: Muchos - 1
    elif len(pivot_params["columns-group"]) > 0 and len(ctp_params["columns-group"]) == 0:
        pivot_df2 = pivot_df.groupby(pivot_params["columns-group"]).agg(
            round(sum(pivot_params["amount-column"]), ROUND_DECIMAL).alias(pivot_params["amount-column"]),
            collect_list(pivot_params["id-column"]).alias(pivot_params["id-column"]))

        ctp_df2 = ctp_df
    # Caso: Muchos - Muchos
    elif len(pivot_params["columns-group"]) > 0 and len(ctp_params["columns-group"]) > 0:
        pivot_df2 = pivot_df.groupby(pivot_params["columns-group"]).agg(
            round(sum(pivot_params["amount-column"]), ROUND_DECIMAL).alias(pivot_params["amount-column"]),
            collect_list(pivot_params["id-column"]).alias(pivot_params["id-column"]))
        ctp_df2 = ctp_df.groupby(ctp_params["columns-group"]).agg(
            round(sum(ctp_params["amount-column"]), ROUND_DECIMAL).alias(ctp_params["amount-column"]),
            collect_list(ctp_params["id-column"]).alias(ctp_params["id-column"]))

    condition = [pivot_df2[col1] == ctp_df2[col2] for col1, col2 in zip(pivot_params["columns-transaction"],
                                                                        ctp_params["columns-transaction"])]
    total_merged = pivot_df2.join(ctp_df2, condition, 'left')
    total_merged = total_merged.withColumn("DIFF", when(col(ctp_params["columns-transaction"][0]).isNotNull(),
                                                        lit(0)).otherwise(lit(None)))
    total_merged = total_merged.select(*pivot_df2.columns, "DIFF")
    condition = [total_merged[col1] == ctp_df2[col2] for col1, col2 in zip(pivot_cols, ctp_cols)]
    merged = total_merged.join(ctp_df2, condition)

    merged = merged.withColumn("DIFF", when(col("DIFF").isNull(),
                               total_merged[pivot_params["amount-column"]] - ctp_df2[ctp_params["amount-column"]]).otherwise(col("DIFF")))

    if len(pivot_params["columns-group"]) == 0 and len(ctp_params["columns-group"]) > 0:
        merged = merged.sort(pivot_params["id-column"])
        merged = merged.dropDuplicates([pivot_cols])
    elif len(pivot_params["columns-group"]) > 0 and len(ctp_params["columns-group"]) == 0:
        merged = merged.sort(ctp_params["id-column"])
        merged = merged.dropDuplicates([ctp_cols])

    merged_df = merged.withColumn("DIFF", round(merged["DIFF"], ROUND_DECIMAL))
    if exclude_pivot:
        df = pivot_df
        group_cols = pivot_params["columns-group"]
        amount_col = pivot_params["amount-column"]
        id_col = pivot_params["id-column"]
    else:
        df = ctp_df
        group_cols = ctp_params["columns-group"]
        amount_col = ctp_params["amount-column"]
        id_col = ctp_params["id-column"]

    total_tmp_cols = group_cols + ["DIFF"]
    df3 = df.join(merged_df.select(*total_tmp_cols), group_cols)

    columns = [col(column) for column in group_cols]

    custom = udf(custom_func_udf, ArrayType(IntegerType()))

    # Fitlrar solo los que tienen S en el campo de exclusión - No tomaria los matches
    # df3 = df3.filter(col(EXCLUDE_ROWS_FIELD) == 'S')

    resultado = df3.groupby(*columns).agg(
        custom(collect_list(amount_col), collect_list(id_col), collect_list(EXCLUDE_ROWS_FIELD), collect_list("DIFF"), lit(max_combinations)).alias("LISTA_DIFF"))

    meged2 = resultado.join(merged_df, group_cols, 'left')

    handle_array_udf = udf(handle_array, ArrayType(IntegerType()))
    meged2 = meged2.withColumn("LISTA_DIFF", handle_array_udf("LISTA_DIFF"))

    meged2 = meged2.filter((col("DIFF") == 0) | ((col("DIFF") != 0) & (size(col("LISTA_DIFF")) > 0)))

    if exclude_pivot:

        meged2 = meged2.withColumn("INTER_PIVOT_ID", array_except(meged2[pivot_params["id-column"]], meged2["LISTA_DIFF"]))
        meged2 = meged2.withColumnRenamed(ctp_params["id-column"], "INTER_CTP_ID")
        if meged2.schema["INTER_CTP_ID"].dataType == LongType():
            meged2 = meged2.withColumn("INTER_CTP_ID", array(col("INTER_CTP_ID")).cast(ArrayType(LongType())))
    else:

        meged2 = meged2.withColumn("INTER_CTP_ID", array_except(meged2[ctp_params["id-column"]], meged2["LISTA_DIFF"]))
        meged2 = meged2.withColumnRenamed(pivot_params["id-column"], "INTER_PIVOT_ID")
        if meged2.schema["INTER_PIVOT_ID"].dataType == LongType():
            meged2 = meged2.withColumn("INTER_PIVOT_ID", array(col("INTER_PIVOT_ID")).cast(ArrayType(LongType())))

    meged2.show()
    print("SHOW:", time.time() - start_time)
    meged2 = meged2.toPandas()
    print("SOLO ALGORITMO:", time.time() - start_time)
    # Guardado en la BD
    print("creando result")
    result = create_result(meged2, descriptor)
    print("emepce a guardar")
    if result["status"] == StatusEnum.ERROR:
        raise InterruptedError(f"Error generando el json resultado. {result['message']}")
    save = save_result(result, session, jdbc_url, descriptor, timezone, pattern)
    if save["status"] == StatusEnum.ERROR:
        raise InterruptedError(f"Error guardando registro resultado en la BD. {result['message']}")

    response["status"] = StatusEnum.OK.name

    return response


def createSession(name: str = "app_engine_spark"):
    try:
        session = SparkSession.builder \
            .appName(name) \
            .getOrCreate()
        return session
    except Exception as e:
        raise Exception(f"Error creando sesion Spark. {e}")


def handle_array(x):
    if isinstance(x, List):
        return x
    else:
        return []


def custom_func_udf(amount_values, id_values, excludes, diffs, max_combinations):
    diff = diffs[0]
    if pd.isna(diff) or diff == 0:
        return None
    diff = int(diff * (10**ROUND_DECIMAL))

    amount_values = [int(value * (10**ROUND_DECIMAL)) for value, exclude in zip(amount_values, excludes) if exclude == 'S']

    dir_name = str(uuid.uuid4())
    prefix = "/tmp/" + dir_name + "_"

    tmp_file_arr1, tmp_file_arr2 = "values.txt", "target.txt"
    full_path_arr1, full_path_arr2 = prefix + tmp_file_arr1, prefix + tmp_file_arr2
    with open(full_path_arr1, 'w') as archivo:
        archivo.writelines([f'{entero}\n' for entero in amount_values])
    with open(full_path_arr2, 'w') as archivo:
        archivo.write(str(diff))
    executable_path = '/tmp/subset_sum_linux'
    indices = []
    for comb in range(1, max_combinations+1):
        argumentos = [full_path_arr1, full_path_arr2, str(comb), '1', '1', 'false', 'false']
        result = subprocess.run([executable_path] + argumentos, check=True, capture_output=True, text=True)
        result = str(result)
        if "keys:[" in result:
            match = result[result.index("keys:[") + 5:result.index("keys remainder") - 20]
            match = match.replace("targets:", "").replace("+", ",")
            match = match.split("==")[0].replace(" ", "")
            match = json.loads(match)
            for idx, val in zip(id_values, amount_values):
                if val in match:
                    indices.append(idx)
                    match.remove(val)
            break
    os.remove(full_path_arr1), os.remove(full_path_arr2)
    return indices


def create_result(data, descriptor):
    result = []
    response = {"detail": result}
    try:
        exclude_pivot = descriptor["config-params"]["exclude-entity-pivot"]
        pivot_params = descriptor["params-input"]["pivot-config"]
        ctp_params = descriptor["params-input"]["counterpart-config"]

        group_pivot_match = pivot_params["columns-group"]
        transaction_pivot_match = pivot_params["columns-transaction"]

        group_counterpart_match = ctp_params["columns-group"]
        transaction_counterpart_match = ctp_params["columns-transaction"]

        used_list = transaction_counterpart_match if exclude_pivot else transaction_pivot_match
        if data is None or data.empty:
            logger.info(f"El dataframe resultado esta vacio")
        else:
            for idx, i in data.iterrows():
                input_data = {}
                key_transaction = None
                key_group_pivot = None
                key_group_counterpart = None
                for element in used_list:
                    if key_transaction is None:
                        key_transaction = str(i[element])
                    else:
                        key_transaction = key_transaction + "-" + str(i[element])
                for element_g in group_pivot_match:
                    if key_group_pivot is None:
                        key_group_pivot = str(i[element_g])
                    else:
                        key_group_pivot = key_group_pivot + "-" + str(i[element_g])
                for element_c in group_counterpart_match:
                    if key_group_counterpart is None:
                        key_group_counterpart = str(i[element_c])
                    else:
                        key_group_counterpart = key_group_counterpart + "-" + str(i[element_c])
                input_data["key-transaction"] = str(key_transaction)
                input_data["key-group-pivot"] = "" if key_group_pivot is None else str(key_group_pivot)
                input_data["key-group-counterpart"] = "" if key_group_counterpart is None else str(
                    key_group_counterpart)
                input_data["list-ids-pivot"] = str(i[FixedFieldsEnum.INTER_PIVOT_ID.value])
                input_data["list-ids-counterpart"] = str(i[FixedFieldsEnum.INTER_CTP_ID.value])
                input_data["exclude-ids"] = str(i[FixedFieldsEnum.LISTA_DIFF.value])
                input_data["difference-amount"] = str(i[FixedFieldsEnum.DIFF.value])
                result.append(input_data)
            response['status'] = StatusEnum.OK
            response["detail"] = result
    except Exception as e:
        logger.error(f"Error al crear el diccionario de resultados. {e}")
        response["status"] = StatusEnum.ERROR
        response["message"] = str(e)
    finally:
        return response


def save_result(result, session, jdbc_url, descriptor, timezone, pattern):
    response = {}
    try:
        d1 = datetime_by_tzone(timezone, pattern)
        result_json = json.dumps(result["detail"])
        data = [descriptor["idScript"], descriptor["idProcess"], d1, FixedFieldsEnum.MATCH_RECORDS.value, result_json]
        df = pd.DataFrame([data], columns=RESULT_TABLE_FIELDS)
        df = session.createDataFrame(df)
        df.write.format("jdbc").option("url", jdbc_url).option("dbtable", RESULT_TABLENAME). \
            option("driver", "com.mysql.cj.jdbc.Driver").mode("append").save()
        response['status'] = StatusEnum.OK
    except Exception as e:
        response["status"] = StatusEnum.ERROR
        response["message"] = str(e)
        logger.error(f"Error al guardar registro en la base de datos {e}")
    finally:
        return response


def datetime_by_tzone(timezone, pattern):
    tzone = timezone
    offset = None
    # Algunos casos donde el timezone es de la forma 4:30 y no se encuentra en timezones de pytz (GMT)
    if ":" in tzone:
        offset = tzone.split(":")[1]
        tzone = tzone.split(":")[0]
        if "+" in tzone:
            tzone = tzone.replace(tzone[-1], str(int(tzone[-1]) + 1))
    timezones_list = pytz.all_timezones
    tzones = [x if tzone in x else None for x in timezones_list]
    tzones = list(filter(None, tzones))
    server_timezone = pytz.timezone(tzones[0])
    logger.debug("Zona Horaria : {}".format(server_timezone))
    server_time = server_timezone.localize(datetime.utcnow())
    current_time = parse(server_time.strftime('%Y-%m-%d %H:%M:%S.%f %Z'))
    if offset:
        offset = pytz.FixedOffset((current_time.utcoffset().total_seconds() / 60 + float(offset)) * -1)
        offset = offset.utcoffset(datetime.utcnow())
        current_time = datetime.utcnow() + offset
    else:
        current_time = current_time.replace(tzinfo=None) - current_time.utcoffset()
    current_time = parse(current_time.strftime(pattern))
    logger.debug("Hora actual: {}".format(current_time))

    return current_time


# Ejecución de proceso
process()
