import json
from typing import Any, Dict
import sys
from wrapt_timeout_decorator import timeout
from app.main.engine.util.EMRServerless import EMRServerless

from app.main.engine.action.ActionInterface import ActionInterface

# RELACION DE IDENTIFICADOR DE ACCION Y NOMBRE DE CLASE
relation_classname_identifier = {
    "match-and-exclude-records-actions": "MatchAndExcludeRecordsAction"
}

# EXCLUDE VALIDATION FIELD
EXCLUDE_ROWS_FIELD = "EXCLUDE_VALID"
# REDONDEO DE DECIMALES
ROUND_DECIMAL = 2


class MatchAndExcludeRecordsAction(ActionInterface):

    def __init__(self, app) -> None:
        super().__init__(app)
        self.descriptor = None
        self.max_combinations = None
        self.timeout = None
        self.exclude_pivot = None
        self.pivot_params = None
        self.ctp_params = None
        self.output = None
        self.config_params = ["max-records-per-combinations", "max-timeout-per-combinations", "exclude-entity-pivot"]

    def parser(self, descriptor: Dict[str, Any]):

        # Validación de parámetros de entrada
        entity_config_params = ["tablename", "id-column", "amount-column", "columns-group", "columns-transaction"]

        if "config-params" not in descriptor.keys():
            raise ReferenceError(f"Parámetro *params-input* no definido en descriptor")

        configs = descriptor["config-params"]
        for param in self.config_params:
            if param not in configs.keys():
                raise ReferenceError(f"Parámetro *{param}* no definido en configuracion")

        if "params-input" not in descriptor.keys():
            raise ReferenceError(f"Parámetro *params-input* no definido en descriptor")

        params_input = descriptor["params-input"]
        if "pivot-config" not in params_input.keys() or "counterpart-config" not in params_input.keys():
            raise ReferenceError(f"Pivot o Contraparte no definidos en descriptor")

        pivot_params, ctp_params = params_input["pivot-config"], params_input["counterpart-config"]
        for param in entity_config_params:
            if param not in pivot_params.keys() or param not in ctp_params.keys():
                raise ReferenceError(f"Parámetro *{param}* no encontrado en pivot o contraparte")

        self.descriptor = descriptor
        self.max_combinations = configs["max-records-per-combinations"]
        self.timeout = configs["max-timeout-per-combinations"]
        self.exclude_pivot = configs["exclude-entity-pivot"]
        self.pivot_params = pivot_params
        self.ctp_params = ctp_params

    def process(self, source_obs, script_name, timezone, pattern):
        try:
            @timeout(self.timeout)
            def __process(source_obj):
                descriptor = DESCRIPTOR
                serverless_job_role_arn = descriptor["job_role_arn"]
                s3_bucket_name = descriptor["bucket_base"]
                search_emr = descriptor["emr_started"]
                end_app = descriptor["terminated_app"]
                emr_serverless = EMRServerless(search_app=search_emr)

                self.app.logger.info("Validando si exite una aplicación ya en curso")
                exists_app = emr_serverless.valid_application()
                if not exists_app["exists"]:
                    self.app.logger.info("Creando e inicializando la aplicación EMR Serverless")
                    emr_serverless.create_application(descriptor["application_name"], descriptor["emr_version"],
                                                      descriptor["app_args"])
                else:
                    emr_serverless.application_id = exists_app["app"]
                emr_serverless.start_application()
                self.app.logger.info(emr_serverless)
                job = descriptor["job"]
                script_location = job["script_location"] + "emr_" + script_name
                jdbc_conn = source_obj.create_spark_connection()
                # jdbc_url = jdbc_conn["url"]
                jdbc_url = "jdbc:mysql://admin:awsadmin@database-2.cgcfmoce13qq.us-east-1.rds.amazonaws.com:3306/cusca"
                # jdbc_properties = jdbc_conn["properties"]
                arguments = [json.dumps(self.descriptor), jdbc_url, timezone, pattern]
                self.app.logger.info("Lanzando nuevo job Spark")
                self.app.logger.info(script_location)
                self.app.logger.info(serverless_job_role_arn)
                self.app.logger.info(arguments)
                self.app.logger.info(job["sparkArgs"])
                self.app.logger.info(s3_bucket_name)
                job_run_id = emr_serverless.run_spark_job(
                    script_location=script_location,
                    job_role_arn=serverless_job_role_arn,
                    arguments=arguments,
                    sparkArguments=job["sparkArgs"],
                    s3_bucket_name=s3_bucket_name,
                )
                job_status = emr_serverless.get_job_run(job_run_id)
                self.app.logger.info(f"Job terminado: {job_run_id}, Estado: {job_status.get('state')}")
                # Fetch and print the logs
                spark_driver_logs = emr_serverless.fetch_driver_log(s3_bucket_name, job_run_id)
                self.app.logger.info("Archivo de salida:\n----\n", spark_driver_logs, "\n----")
                if end_app:
                    # Now stop and delete your application
                    self.app.logger.info("Deteniendo y borrando aplicación Spark")
                    emr_serverless.stop_application()
                    emr_serverless.delete_application()
                self.app.logger.info("Hecho! 👋")

        except TimeoutError as e:
            raise TimeoutError(f"Tiempo límite superado. {e}")

        self.output = __process(source_obs)

    def response(self):
        return self.output


DESCRIPTOR = {
  "application_name": "css_cuscatlan",
  "emr_version": "emr-7.0.0",
  "emr_started": True,
  "terminated_app": False,
  "job_role_arn": "arn:aws:iam::000026703603:role/emr-serverless-job-role",
  "bucket_base": "bucket-emr-example",
  "app_args": {
    "initial_capacity": {
        "DRIVER": {
            "workerCount": 1,
            "workerConfiguration": {
                "cpu": "16vCPU",
                "memory": "32GB"
            }
        },
        "EXECUTOR": {
            "workerCount": 12,
            "workerConfiguration": {
                "cpu": "16vCPU",
                "memory": "32GB"
            }
        }
    },
    "maximun_capacity": {
        "cpu": "208vCPU",
        "memory": "416GB",
        "disk": "1000GB"
    },
    "imageConfiguration": {
      "imageUri": "000026703603.dkr.ecr.us-east-1.amazonaws.com/css_spark_custom:0.0.5"
    },
    "networkConfiguration": {
      "subnetIds": [
        "subnet-0f86499848ec99861", "subnet-078fe716da8b53818", "subnet-0a7d0a8bc3b623474"
      ],
      "securityGroupIds": [
        "sg-02154713a3639f7ce"
      ]
    }
  },
  "job": {
    "script_location": "s3://bucket-emr-example/css_cusca/endpoint/",
    "sparkArgs": {
      "driver-cores": 16,
      "driver-memory": "14g",
      "executor-cores": 16,
      "executor-memory": "14g",
      "executor-instances": 12,
      "others": "--jars s3://bucket-emr-example/bcom_spark/jars/mysql-connector-java-8.0.30.jar"
    }
  }
}
